/* Decompress a memory region */

#ifdef LITTLE_ENDIAN
# define fixw(i)	(i = ((unsigned short)i>>8) | \
				(((unsigned short)i & 0xFF) << 8))
# define fixl(i)	(i = (typeof(i))(((unsigned long)i>>24) | \
				((((unsigned long)i>>16) & 0xFF) << 8) | \
				((((unsigned long)i>>8) & 0xFF) << 16) | \
				(((unsigned long)i & 0xFF) << 24)))
#else
# define fixw(i)
# define fixl(i)
#endif

struct hdr {
    short hdrlen;
    short method;
    long fullsize;	/* size of uncompressed data */
    long compsize;	/* size of compressed data excl. header */
};

long decompress(register void *src, register void *dest, register long destsize)
{
    register struct hdr *hdr = src;
    register short *s = (short *)((char *)src + sizeof(struct hdr));
    register long size;
    register short *end;
    register long cnt;

    fixw(hdr->hdrlen);
    fixw(hdr->method);
    fixl(hdr->fullsize);
    fixl(hdr->compsize);

    size = hdr->compsize;
    end = s + size/sizeof(short);

    if (hdr->fullsize > destsize)
	return (hdr->fullsize - destsize) | 0x80000000;

    switch (hdr->method) {
    case 0:	/* No compression */
	bcopy(s, dest, size);
	break;

    case 1: {	/* Word RLC */
	short *d = dest, val;

	while (s < end) {
	    val = *s++;
	    fixw(val);
	    if (val < 0) {
		if (val == -0x8000) {
		    cnt = *(long *)s;
		    s += 2;
		    fixl(cnt);
		} else
		    cnt = -val;
		val = *s++;
		while (cnt--)
		    *d++ = val;
	    } else {
		/* Store the original value, not the endian-converted one;
		 * leave data fixing to the caller */
		*d++ = s[-1];
	    }
	}
	break;
    } /* case 1 */

    case 2: {	/* Long RLC */
	long *d = dest, val;
	short sval;

	while (s < end) {
	    sval = *s;
	    fixw(sval);
	    if (sval < 0) {
		s++;
		if (sval == -0x8000) {
		    cnt = *(long *)s;
		    s += 2;
		    fixl(cnt);
		} else
		    cnt = -sval;
		val = *(long *)s;
		s += 2;
		while (cnt--)
		    *d++ = val;
	    } else {
		*d++ = *(long *)s;
		s += 2;
	    }
	}
	break;
    } /* case 2 */

    } /* switch (hdr->method) */

    return 0;
}
