/* Routines for handling maps */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "map.h"

/* For little-endian machines we have to kluge around a lot because the
 * files are stored in big-endian order.  But then, little-endian machines
 * are stupid anyway. :) */
#ifdef LITTLE_ENDIAN
# define fixw(i)	(i = ((unsigned short)i>>8) | \
				(((unsigned short)i & 0xFF) << 8))
# define fixl(i)	(i = (typeof(i))(((unsigned long)i>>24) | \
				((((unsigned long)i>>16) & 0xFF) << 8) | \
				((((unsigned long)i>>8) & 0xFF) << 16) | \
				(((unsigned long)i & 0xFF) << 24)))
#endif


struct TileData *m_TileSet = NULL;
char m_TileFile[256];

#ifndef TILES_ONLY
struct Map *m_CurMap = NULL;
short m_CurScrNum, m_CurLayer;
#endif

static int LoadLayers(FILE *, struct MapScrData *);


struct TileData *LoadTileSet(char *file2)
{
    FILE *f;
    struct TileData *td;
    short ID;
    int i;

    if (f = fopen(file2, "r")) {
	if (td = malloc(sizeof(*td))) {
	    fread(&ID, 2, 1, f);
#ifdef LITTLE_ENDIAN
	    fixw(ID);
#endif
	    if (ID < 0x1000) {
		short rgb[32], i;

		td->ID = TILEID_T16;
		fseek(f, 0, SEEK_SET);
		fread(rgb, 2, 32, f);
		for (i = 0; i < 32; ++i) {
		    short c = rgb[i];
#ifdef LITTLE_ENDIAN
		    fixw(c);
#endif
		    td->RGB[i] = (c&0xF00)*0x11<<8 | (c&0xF0)*0x11<<4 |
								(c&0xF)*0x11;
		}
		for (i = 32; i < 256; ++i)
		    td->RGB[i] = 0;
		fread(td->Tiles, sizeof(td->Tiles), 1, f);
#ifdef LITTLE_ENDIAN
		for (i = 0; i < 256; ++i) {
		    short j;
		    for (j = 0; j < 128; ++j) fixw(td->Tiles[i][j]);
		}
#endif
		fread(td->Comments, sizeof(td->Comments), 1, f);
		fread(td->Groups, sizeof(td->Groups), 1, f);
#ifdef LITTLE_ENDIAN
		for (i = 0; i < 256; ++i) {
		    short j;
		    for (j = 0; j < 4; ++j) fixw(td->Groups[i][j]);
		}
#endif
		fclose(f);
		return td;
	    }
	    switch (ID) {
	    case TILEID_T16:
		fread(td->RGB, ((char *)(&td[1])-(char *)td->RGB), 1, f);
#ifdef LITTLE_ENDIAN
		for (i = 0; i < 256; ++i) {
		    short j;
		    fixl(td->RGB[i]);
		    for (j = 0; j < 128; ++j) fixw(td->Tiles[i][j]);
		    for (j = 0; j < 4; ++j) fixw(td->Groups[i][j]);
		}
#endif
		fclose(f);
		return td;
	    default:
		free(td);
		fclose(f);
	    }
	} else {
	    fclose(f);
	}
    }
    return NULL;
}

void FreeTileSet(struct TileData *td)
{
    free(td);
}


#ifndef TILES_ONLY

/* Load a map's screen's layers. */
static int LoadLayers(FILE *f, struct MapScrData *sd)
{
    int i, comp, nlayers = sd->NumLayers, size = sd->Width * sd->Height * 2;
    struct LayerData *lp = sd->LayerData;
    void *compmem;
    long j, *ptr2;
    short *ptr;

    for (i = 0; i < nlayers; i++, lp++) {
	comp = (lp->CompSize & 0x80000000) != 0;
	lp->CompSize &= 0x7FFFFFFF;
	fseek(f, (long)lp->Data, SEEK_SET);
	if (!(lp->Data = malloc(lp->CompSize))) {
	    for (lp--; i; i--, lp--)
		free(lp->Data);
	    return 0;
	}
	fread(lp->Data, lp->CompSize, 1, f);
	if (comp) {
	    if (!(compmem = malloc(size))) {
		free(lp->Data);
		for (lp--; i; i--, lp--)
		    free(lp->Data);
		return 0;
	    }
	    if (decompress(lp->Data, compmem, size) & 0x80000000) {
		free(compmem);
		free(lp->Data);
		for (lp--; i; i--, lp--)
		    free(lp->Data);
		return 0;
	    }
	    free(lp->Data);
	    lp->Data = compmem;
	}
#ifdef LITTLE_ENDIAN
	ptr = lp->Data;
	for (j = size/2; j; --j) {
	    fixw(*ptr); ++ptr;
	}
#endif
    }
    return 1;
}

/* Load a map.  If replace, then replace m_CurMap etc. with new screen.
 * If !replace, then store pointer to tile set in *tdptr (unless it's
 * NULL).
 */
struct Map *LoadMap(char *file2, int replace, struct TileData **tdptr)
{
    FILE *f;
    void *compmem;
    struct Map *map;
    struct MapScrData *sd = NULL;
    long a, i;
    struct TileData *td;
    short id, rev;
    long buf[32];
    struct LayerData *lp;
    char s[256], *t;
    short rgb[32];

    if (f = fopen(file2, "r")) {
	if (map = malloc(sizeof(struct Map))) {
	    fread(map, 2, 1, f);	/* read ID */
#ifdef LITTLE_ENDIAN
	    fixw(map->ID);
#endif
	    id = map->ID & 0xFF00; rev = map->ID & 0x00FF;
	    switch (id) {
	    case MAPID_T16:
		if (rev < 4) {
		    if (rev < 2) {
			fread((void *)((long)map+2), 254, 1, f);
			fread(rgb, 64, 1, f);
			fread((void *)((long)map+256), sizeof(*map)-256, 1, f);
		    } else
			fread((void *)((long)map+2), sizeof(*map)-2, 1, f);
#ifdef MAP_ALIGN4
		    fseek(f, -6, SEEK_CUR);
		    fread(&map->Screens, 4, 1, f);
#endif
#ifdef LITTLE_ENDIAN
		    fixw(map->NumScreens); fixl(map->Screens);
#endif
		    /* Allocate memory for screen pointers */
		    if (!(compmem = malloc(map->NumScreens*4))) {
			free(map);
			fclose(f);
			return NULL;
		    }
		    fseek(f, (long)map->Screens, SEEK_SET);
		    fread(map->Screens = compmem, map->NumScreens*4, 1, f);
#ifdef LITTLE_ENDIAN
		    for (i = 0; i < map->NumScreens; ++i)
			fixl(map->Screens[i]);
#endif
		    for (a = 0; a < map->NumScreens; a++) {
			if (!(sd = calloc(sizeof(struct MapScrData), 1)))
			    goto sd_nomem;
			fseek(f, (long)map->Screens[a], SEEK_SET);
			map->Screens[a] = sd;
			if (rev < 3) {
			    fread(sd, 12, 1, f);
			    if (rev < 2) {
				for (i = 0; i < 32; i++) {
				    short c = rgb[i];
#ifdef LITTLE_ENDIAN
				    fixw(c);
#endif
				    sd->ScrRGB[i] =
						((c & 0xF00) << 12) |
						((c & 0x0F0) <<  8) |
						((c & 0x00F) <<  4);
				}
			    } else {
				fread(sd->ScrRGB, 4, 32, f);
#ifdef LITTLE_ENDIAN
				for (i = 0; i < 32; ++i)
				    fixl(sd->ScrRGB[i]);
#endif
			    }
			} else {
			    fread(sd, sizeof(*sd), 1, f);
#ifdef LITTLE_ENDIAN
			    for (i = 0; i < 256; ++i)
				fixl(sd->ScrRGB[i]);
#endif
			}
#ifdef LITTLE_ENDIAN
			fixw(sd->ScrID); fixw(sd->NumLayers);
			fixw(sd->Width); fixw(sd->Height);
			fixl(sd->LayerData);
#endif
			lp = malloc(sd->NumLayers*sizeof(struct LayerData));
			if (!lp) goto lp_nomem;
			if (rev == 0) {
			    fread(buf, 0x80, 1, f);
			    for (i = 0; i < sd->NumLayers; i++) {
				lp[i].CompSize = buf[i];
				lp[i].Data = (void *)buf[i+16];
			    }
			} else {
			    fseek(f, (long)sd->LayerData, SEEK_SET);
			    fread(lp, sd->NumLayers*sizeof(struct LayerData),
									1, f);
			}
			sd->LayerData = lp;
#ifdef LITTLE_ENDIAN
			for (i = 0; i < sd->NumLayers; ++lp, ++i) {
			    fixl(lp->CompSize); fixl(lp->Data);
			}
#endif
			if (!LoadLayers(f, sd)) goto lp_nomem;
		    }
		    fclose(f);
		    if (m_CurMap && replace)
			FreeMap(m_CurMap);
		    if (strcmp(m_TileFile,map->TileFile) ||
						(!replace && tdptr)) {
			if (m_TileSet && replace)
			    FreeTileSet(m_TileSet);
			if (!(td = LoadTileSet(map->TileFile))) {
			    strcpy(s, file2);
			    if ((t = strrchr(s, '/')) || (t = strrchr(s, ':')))
				t[1] = 0;
			    else
				*s = 0;
			    if ((t = strrchr(map->TileFile, '/')) ||
					    (t = strrchr(map->TileFile, ':')))
				strcat(s, t+1);
			    else
				strcat(s, map->TileFile);
			    if (!(td = LoadTileSet(map->TileFile))) {
				if ((t = strrchr(map->TileFile, '/')) ||
					    (t = strrchr(map->TileFile, ':')))
				    strcpy(s, t+1);
				else
				    strcpy(s, map->TileFile);
				if (!(td = LoadTileSet(map->TileFile))) {
				    FreeMap(map);
				    return NULL;
				}
			    }
			}
		    } else
			td = m_TileSet;
		    if (replace) {
			m_CurMap = map;
			m_TileSet = td;
			strcpy(m_TileFile, map->TileFile);
		    } else
			if (tdptr) *tdptr = td;
		    return map;
		    break;
		} /* else fall through */

	    default:
		free(map);
		fclose(f);
		return NULL;
	    } /* switch (ID) */
	} else /* couldn't allocate map */
	    fclose(f);
    } /* if (fopen()) */
    return NULL;

lp_nomem:
    free(sd);
sd_nomem:
    for (a--; a >= 0; a--) FreeMapScreen(map, a);
    free(map->Screens);
    free(map);
    fclose(f);
    return NULL;
}

struct Map *LoadMapName(char *file)
{
    return LoadMap(file, 1, NULL);
}


#ifndef MAP_NODISPLAY

void SetMapDisp(struct Map *map, struct TileData *td)
{
    m_CurMap = map; m_TileSet = td;
}

void DispMap(short id, short layer, struct Screen *scr,
			short x0, short y0, short tlx, short tly,
			short w, short h, short x, short y)
{
    struct MapScrData *sd;
    register int i = 0;

    sd = m_CurMap->Screens[m_CurScrNum];
    if (id != sd->ScrID) {
	while ((sd = m_CurMap->Screens[++i])->ScrID != id);
	m_CurScrNum = i;
    }
    m_CurLayer = layer;
    DispMap16(sd->LayerData[layer].Data, (short *)m_TileSet->Tiles,
		    scr, x0, y0, tlx, tly, SWidth, SHeight, w, h, x, y);
}

#endif	/* MAP_NODISPLAY */


void FreeMapScreen(struct Map *map, int n)
{
    int j;
    struct MapScrData *sd = map->Screens[n];

    for (j = 0; j < sd->NumLayers; j++)
	free(sd->LayerData[j].Data);
    free(sd->LayerData);
    free(sd);
}

void FreeMap(struct Map *map)
{
    int i;

    for (i = 0; i < map->NumScreens; i++) FreeMapScreen(map, i);
    free(map->Screens);
    free(map);
}

#endif	/* TILES_ONLY */

void MapCleanup()
{
#ifndef TILES_ONLY
    if (m_CurMap) FreeMap(m_CurMap);
#endif
    if (m_TileSet) FreeTileSet(m_TileSet);
}
