#ifndef MAP_H
#define MAP_H

/* Structures & defines for tile sets and maps */

/* Defines that can be used with map.[ch]:
 *	MAP_NODISPLAY	Exclude display routines from compilation.
 *	MAP_ALIGN4	Define if your compiler forces 4-byte alignment of
 *			   pointers and/or longs no matter what.  NOTE:
 *			   maps read in with this define active should NOT
 *			   be written back to disk, as they will not be
 *			   readable!
 *	LITTLE_ENDIAN	Compile on a little-endian (least significant byte
 *			   first) machine.  Required because the files are
 *			   in big-endian format.
 */
extern struct Map *m_CurMap;
extern short m_CurScrNum,m_CurLayer;  /* CurScrNum is Screens[] index */
extern struct TileData *m_TileSet;
extern char m_TileFile[];

struct TileData {	/*/ td_ */
    short ID;
    long RGB[256];
    short Tiles[256][128];
    char Comments[256][64];
    long Groups[256][4];
};

#define TILEID_T16	0x1000

struct Map {		/*/ m_ */
    short ID;
    char  TileFile[254];
    short NumScreens;
#ifdef MAP_ALIGN4
    short pad;
#endif
    struct MapScrData **Screens;
};

#define MAPID_T8	0x0100	/* obsolete */
#define MAPID_T16	0x0200

struct LayerData {	/*/ ld_ */
    long CompSize;
    short *Data;
};

struct MapScrData {	/*/ msd_ */
    short ScrID;
    short NumLayers;
    short Width, Height;
    struct LayerData *LayerData;
    long ScrRGB[256];
};

/* Macros for easy access to structures */

#ifndef NO_MAP_MACROS
#define TileFilename m_CurMap->TileFile
#define NScr    m_CurMap->NumScreens
#define NLayers m_CurMap->Screens[m_CurScrNum]->NumLayers
#define SWidth  m_CurMap->Screens[m_CurScrNum]->Width
#define SHeight m_CurMap->Screens[m_CurScrNum]->Height
#define CSize   m_CurMap->Screens[m_CurScrNum]->LayerData[m_CurLayer].CompSize
#define ScrData m_CurMap->Screens[m_CurScrNum]->LayerData[m_CurLayer].Data
#define SRGB    m_CurMap->Screens[m_CurScrNum]->ScrRGB
#endif

#define GetSTile(x,y) ScrData[(y)*SWidth+(x)]
#define PutSTile(x,y,c) ScrData[(y)*SWidth+(x)]=c

#ifndef MAP_NODISPLAY
void DispMap16(short *, short *, struct Screen *,
	short, short, short, short, short, short, short, short, short, short);
/* Enough parameters for you? :-) */
#endif

extern struct TileData *LoadTileSet(char *);
extern void FreeTileSet(struct TileData *);
extern struct Map *LoadMap(char *, int, struct TileData **);
extern struct Map *LoadMapName(char *);
#ifndef MAP_NODISPLAY
extern void SetMapDisp(struct Map *, struct TileData *);
extern void DispMap(short, short, struct Screen *,
	short, short, short, short, short, short, short, short);
#endif
extern void FreeMapScreen(struct Map *, int);
extern void FreeMap(struct Map *);
extern void MapCleanup(void);

#endif
