/* Convert a map to a GIF image */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "map.h"
#include "options.h"
#include "tile.h"

#ifdef LITTLE_ENDIAN         
# define fixw(i)	(i = ((unsigned short)i>>8) | \
				(((unsigned short)i & 0xFF) << 8))
# define fixl(i)	(i = (typeof(i))(((unsigned long)i>>24) | \
				((((unsigned long)i>>16) & 0xFF) << 8) | \
				((((unsigned long)i>>8) & 0xFF) << 16) | \
				(((unsigned long)i & 0xFF) << 24)))
#else
# define fixw(i)
# define fixl(i)
#endif
                        
extern long decompress(void *src, void *dest, long destsize);
extern void copy1(unsigned short tiles[][128], short *data, short y,
								pixel *dest);
extern void copy2(unsigned short tiles[][128], short *data, short y,
								pixel *dest);
extern void copy4(unsigned short tiles[][128], short *data, short y,
								pixel *dest);
extern void copy8(unsigned short tiles[][128], short *data, short y,
								pixel *dest);
extern void copy16(unsigned short tiles[][128], short *data, short y,
								pixel *dest);

short planes;

main(int ac, char **av)
{
    struct TileData *tiles;
    struct Map *map;
    FILE *out;
    register short *src, y, x, w, h, i, scale, scale2, width_row;
    register short *tileptr;
    unsigned char *row, *rptr;
    long size_row;
    static int planes_tmp = 8, scale_tmp = 1, screen = 0, layer = 0;

    option optlist[] = {
	{ "depth", OPT_INT, &planes_tmp },
	{ "scale", OPT_INT, &scale_tmp },
	{ "screen", OPT_INT, &screen },
	{ "layer", OPT_INT, &layer },
	{ NULL }
    };

    ac = do_options(ac, av, optlist);
    planes = planes_tmp; scale = scale_tmp;

    scale2 = 16 / scale;

    if (ac != 3) {
	fprintf(stderr, "\
Converts a map file to a GIF image.\n\
Usage: %s input-map output-image [options]\n\
Options:\n\
	-depth      Specify depth of destination image (default 8)\n\
	-scale      Specify scaling factor; '-scale n' shrinks image by a\n\
			factor of 'n'\n\
	-screen     Specify which screen ID in the map to use (default 0)\n\
	-layer      Specify which layer in the screen to use (default 0)\n",
		av[0]);
	exit(10);
    }
    if (scale != 1 && scale != 2 && scale != 4 && scale != 8 && scale != 16) {
	fprintf(stderr, "Scale factor must be 1, 2, 4, 8, or 16\n");
	exit(20);
    }
    if (!(map = LoadMap(av[1], 0, &tiles))) {
	fprintf(stderr, "Can't load map\n");
	exit(20);
    }

    w = map->Screens[screen]->Width; h = map->Screens[screen]->Height;

    /* Size of a row:
     *    'w' tiles * 'scale2^2' pixels per tile * 1 byte per pixel
     */
    width_row = w*scale2;
    row = malloc(size_row = width_row*scale2);
    if (!row) {
	fprintf(stderr,"Out of memory: couldn't allocate %d bytes\n",size_row);
	exit(20);
    }

    colorsinmainmap = 1<<planes;
    for (i = 0; i < (1<<planes); ++i) {
	long c = map->Screens[screen]->ScrRGB[i];
	MainColorMap[0][i] = c>>16;
	MainColorMap[1][i] = (c>>8) & 0xFF;
	MainColorMap[2][i] = c & 0xFF;
    }
    GifScreen.Width = w * scale2;
    GifScreen.Height = h * scale2;

    if (!fopen_gif_file(av[2], "w")) {
	fprintf(stderr, "Cannot open output file\n");
	exit(20);
    }

    src = (short *)map->Screens[screen]->LayerData[layer].Data;
    tileptr = tiles->Tiles;
    for (y = 0; y < h; ++y) {
	for (rptr = row, x = 0; x < w; rptr += scale2, ++x) {
	    for (i = 0; i < scale2; ++i) {
		switch (scale) {
		    case 1: copy1(tileptr, src, i, rptr + i*width_row); break;
		    case 2: copy2(tileptr, src, i*2, rptr + i*width_row); break;
		    case 4: copy4(tileptr, src, i*4, rptr + i*width_row); break;
		    case 8: copy8(tileptr, src, i*8, rptr + i*width_row); break;
		    case 16: copy16(tileptr, src, i*16, rptr + i*width_row); break;
		}
	    }
	    ++src;
	}
	for (i = 0; i < scale2; ++i) {
	    if (!write_gif_line(row + i*w*scale2)) {
		perror("Write failed");
		exit(20);
	    }
	}

#if 0
	if (!(y & 31)) {
	    fprintf(stderr, "%3d%% complete\r", (y*100) / h); fflush(stderr);
	}
#endif
    }

#if 0
    printf("100%% complete\n");
#endif
    fclose_gif_file(out);
    exit(0);
}
