/* Convert a map to an IFF ILBM image */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "map.h"
#include "options.h"

#ifdef LITTLE_ENDIAN         
# define fixw(i)	(i = ((unsigned short)i>>8) | \
				(((unsigned short)i & 0xFF) << 8))
# define fixl(i)	(i = (typeof(i))(((unsigned long)i>>24) | \
				((((unsigned long)i>>16) & 0xFF) << 8) | \
				((((unsigned long)i>>8) & 0xFF) << 16) | \
				(((unsigned long)i & 0xFF) << 24)))
#endif
                        
extern long decompress(void *src, void *dest, long destsize);
extern short copy2(short *tiles, short *data, short y, short p);
extern short copy4(short *tiles, short *data, short y, short p);
extern short copy8(short *tiles, short *data, short y, short p);
extern short copy16(short *tiles, short *data, short y, short p);

main(int ac, char **av)
{
    struct TileData *tiles;
    struct Map *map;
    FILE *out;
    long buf[12];
    register short *src, *rptr, y, x, w, h, planes,
	i, j, scale, scale2, w2, width_row, x2;
    register short *tils;
    short *row;
    long size_row;
    static int planes_tmp = 8, scale_tmp = 1, screen = 0, layer = 0;

    option optlist[] = {
	{ "depth", OPT_INT, &planes_tmp },
	{ "scale", OPT_INT, &scale_tmp },
	{ "screen", OPT_INT, &screen },
	{ "layer", OPT_INT, &layer },
	{ NULL }
    };

    ac = do_options(ac, av, optlist);
    planes = planes_tmp; scale = scale_tmp;

    scale2 = 16 / scale;

    if (ac != 3) {
	fprintf(stderr, "\
Converts a map file to an uncompressed IFF ILBM image.\n\
Usage: %s input-map output-image [options]\n\
Options:\n\
	-depth      Specify depth of destination image (default 8)\n\
	-scale      Specify scaling factor; '-scale n' shrinks image by a\n\
			factor of 'n'\n\
	-screen     Specify which screen ID in the map to use (default 0)\n\
	-layer      Specify which layer in the screen to use (default 0)\n",
		av[0]);
	exit(10);
    }
    if (scale != 1 && scale != 2 && scale != 4 && scale != 8 && scale != 16) {
	fprintf(stderr, "Scale factor must be 1, 2, 4, 8, or 16\n");
	exit(20);
    }
    if (!(map = LoadMap(av[1], 0, &tiles))) {
	fprintf(stderr, "Can't load map\n");
	exit(20);
    }

    w = map->Screens[screen]->Width; h = map->Screens[screen]->Height;
    w2 = (w + scale-1) & ~(scale-1);  /* so that a row is word-aligned */

    /* Size of a row:
     *    'w2' tiles * 'scale2^2' pixels per tile * 'planes' planes per pixel
     *    -------------------------------------------------------------------
     *                             8 pixels per byte
     */
    row = malloc(size_row = w2*scale2*scale2*planes/8);
    width_row = w2*scale2/16; /* in shorts */
    if (!row) {
	fprintf(stderr,"Out of memory: couldn't allocate %d bytes\n", size_row);
	exit(20);
    }

    out = fopen(av[2], "w");
    if (!out) {
	fprintf(stderr, "Cannot open output file\n");
	exit(20);
    }

    buf[0] = 'FORM';
    buf[1] = 48 + (3<<planes) + h*size_row;
    buf[2] = 'ILBM';
    buf[3] = 'BMHD';
    buf[4] = 20;
    buf[5] = ((w*scale2)<<16) | (h*scale2);
    buf[6] = 0;
    buf[7] = planes<<24;
    buf[8] = 0x00000101;
    buf[9] = ((w*16)<<16) | (h*16);
    buf[10] = 'CMAP';
    buf[11] = 3<<planes;
#ifdef LITTLE_ENDIAN
    for (i = 0; i < 12; ++i)
	fixl(buf[i]);
#endif
    fwrite(buf, 4, 12, out);
    for (i = 0; i < 1<<planes; ++i) {
	long c = map->Screens[screen]->ScrRGB[i];
	fputc(c>>16, out);
	fputc((c>>8) & 0xFF, out);
	fputc(c & 0xFF, out);
    }
    buf[0] = 'BODY';
    buf[1] = h*size_row;
#ifdef LITTLE_ENDIAN
    fixl(buf[0]); fixl(buf[1]);
#endif
    fwrite(buf, 4, 2, out); fflush(out);

    src = (short *)map->Screens[screen]->LayerData[layer].Data;
    tils = tiles->Tiles;
    for (y = 0; y < h; ++y) {
	for (x = x2 = 0; x < w; ++x2, x += scale) {
	    rptr = row + x2;
	    for (i = 0; i < scale2; ++i) {
		for (j = 0; j < planes; ++j) {
		    switch (scale) {
			case 1: *rptr = tils[(*src)*128 + i*8 + j]; break;
			case 2: *rptr = copy2(tils, src, i, j); break;
			case 4: *rptr = copy4(tils, src, i, j); break;
			case 8: *rptr = copy8(tils, src, i, j); break;
			case 16: *rptr = copy16(tils, src, i, j); break;
		    }
#ifdef LITTLE_ENDIAN
		    fixw(*rptr);
#endif
		    rptr += width_row;
		}
	    }
	    if (w-x < scale)
		src += w-x;
	    else
		src += scale;
	}
	if (fwrite(row, size_row, 1, out) <= 0) {
	    perror("Write failed");
	    exit(20);
	}

#if 0
	if (!(y & 31)) {
	    fprintf(stderr, "%3d%% complete\r", (y*100) / h); fflush(stderr);
	}
#endif
    }

#if 0
    printf("100%% complete\n");
#endif
    fclose(out);
    exit(0);
}
