#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define strscpy(d,s,n) do { strncpy(d, s, n); d[(n)-1] = 0; } while (0)
#define lenof(a) (sizeof(a)/sizeof(*(a)))

/* This should be the path to the directory containing the map/tile data
 * files. */
#ifndef MAPPATH
# define MAPPATH "/home/achurch/ff1-maps"
#endif

/* This should be the path to the map2gif program. */
#ifndef MAP2GIF
# define MAP2GIF "/home/achurch/ff1-maps/map2gif"
#endif

/* This is the format for the temporary file (%d is replaced by the PID). */
#undef TMPFORMAT
#define TMPFORMAT "/tmp/map2gif.%d"

/*************************************************************************/

/* List of locations.  Used both to sanity-check input and to add links to
 * generated HTML. */

struct location {
    const char *name, *title;
    const char *floors[8];
} locations[] = {
    {"Coneria",         "Town of Coneria",  {"1F"}},
    {"ConeriaCastle",   "Coneria Castle",   {"1F","2F"}},
    {"TempleOfFiends",  "Temple of Fiends", {"B1"}},
    {"MatoyasCave",     "Matoya's Cave",    {"B1"}},
    {"Pravoka",         "Town of Pravoka",  {"1F"}},
    {"Elfland",         "Elfland",          {"1F"}},
    {"ElvenCastle",     "Elven Castle",     {"1F"}},
    {"NorthwestCastle", "Northwest Castle", {"1F"}},
    {"MarshCave",       "Marsh Cave",       {"B1","B2","B3"}},
    {"DwarfCave",       "Dwarf Cave",       {"B1"}},
    {"Melmond",         "Town of Melmond",  {"1F"}},
    {"TitansPass",      "Titan's Pass",     {"B1"}},
    {"SardasCave",      "Sarda's Cave",     {"B1"}},
    {"EarthCave",       "Earth Cave",       {"B1","B2","B3","B4","B5"}},
    {"CrescentLake",    "Crescent Lake",    {"1F"}},
    {"GurguVolcano",    "Gurgu Volcano",    {"B1","B2","B3","B4","B5"}},
    {"IceCave",         "Ice Cave",         {"B1","B2","B3"}},
    {"CardiaIslands",   "Cardia Islands",   {"B1A","B1B","B2"}},
    {"CastleOfOrdeal",  "Castle of Ordeal", {"1F","2F","3F"}},
    {"Onrac",           "Town of Onrac",    {"1F"}},
    {"SeaShrine",       "Sea Shrine",       {"B1","B2","B3","B4","B5"}},
    {"Waterfall",       "Waterfall Cave",   {"B1"}},
    {"Gaia",            "Town of Gaia",     {"1F"}},
    {"Lefein",          "Town of Lefein",   {"1F"}},
    {"MirageTower",     "Mirage Tower",     {"1F","2F","3F"}},
    {"FloatingCastle",  "Floating Castle",  {"1F","2F","3F","4F","5F"}},
    {"TempleOfFiends2", "Temple of Fiends Revisited",
	{"1F","2F","3F","B1","B2","B3","B4","B5"}},
    {NULL}
};

/*************************************************************************/
/*************************************************************************/

/* An error occurred - print an error message and quit */
void doerror(char *s)
{
    printf("Content-Type: text/html\n\n"
		"<html>\n<head>\n<title>Error</title>\n</head>\n"
		"<body>\n<h1>Error!</h1>\n"
		"Something has gone horribly wrong...<p>\n"
		"%s<p>\n</body>\n</html>\n", s);
    exit(20);
}

/*************************************************************************/

/* Translate a 2-digit hex number to a character */
char dohexnum(char *s)
{
    char t[3];
    strncpy(t, s, 2); t[2] = 0;

    if (!isxdigit(s[0]) || !isxdigit(s[1])) {
	doerror("Bad hex digit");
    }
    return strtol(t, NULL, 16);
}

/*************************************************************************/

/* Return the struct location * for the given map name (NULL if not found). */

struct location *find_location(const char *name)
{
    struct location *loc;

    for (loc = locations; loc->name != NULL; loc++) {
	if (strcmp(loc->name, name) == 0)
	    return loc;
    }
    return NULL;
}

/*************************************************************************/
/*************************************************************************/

int send_html(struct location *loc, int screen, int layer, int scale)
{
    int i;

    printf("Content-Type: text/html; charset=iso-8859-1\r\n\r\n");
    printf("\
<html>\
<head>\
<title>Final Fantasy 1: %s %s</title>\
</head>\
<body>\
<div align=center>\
<h2>%s %s</h2>\
<p>Floor: ",
	   loc->title, loc->floors[screen], loc->title, loc->floors[screen]);
    for (i = 0; i < lenof(loc->floors) && loc->floors[i]; i++) {
	if (i > 0)
	    printf(" | ");
	if (i != screen)
	    printf("<a href=\"getmap.cgi?loc=%s/%d&scale=%d\">",
		   loc->name, i, scale);
	printf("%s", loc->floors[i]);
	if (i != screen)
	    printf("</a>");
    }
    printf("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Scale: ");
    for (i = 1; i <= 8; i <<= 1) {
	if (i > 1)
	    printf(" | ");
	if (i != scale) {
	    if (loc->floors[1])
		printf("<a href=\"getmap.cgi?loc=%s/%d&scale=%d\">",
		       loc->name, screen, i);
	    else
		printf("<a href=\"getmap.cgi?loc=%s&scale=%d\">",
		       loc->name, i);
	}
	printf("1 : %d", i);
	if (i != scale)
	    printf("</a>");
    }
    printf("\
<p><img src=\"getmap.cgi?loc=%s/%d&scale=%d&gif=1\">\
<p><a href=\"/ff1-maps.html\">Return to location list</a>\
<p><hr width=\"50%\">\
<p><i>Final Fantasy is copyright (c) SQUARE ENIX Co., Ltd.</i>\
</div>\
</body>\
</html>",
	   loc->name, screen, scale);
    return 0;
}

/*************************************************************************/

int send_gif(struct location *loc, int screen, int layer, int scale)
{
    static char tmp[1024], s[1024];
    int i;
    FILE *f;

    if (chdir(MAPPATH) < 0) {
	doerror("Can't chdir to map directory\n");
	return 1;
    }
    sprintf(tmp, TMPFORMAT, getpid());

    sprintf(s, "%s %s %s -depth 4 -scale %d", MAP2GIF, loc->name, tmp, scale);
    if (screen >= 0)
	sprintf(s+strlen(s), " -screen %d", screen);
    if (layer >= 0)
	sprintf(s+strlen(s), " -layer %d", layer);
    system(s);

    f = fopen(tmp, "r");
    if (!f) {
	printf("Content-Type: text/plain\n\noops, error\n");
	i = -1;
    } else {
	printf("Content-Type: image/gif\n\n");
	/* Unlink the temp file now, to make sure it disappears when we
	 * exit whether normally, by SIGKILL, or whatever. */
	unlink(tmp);
	while ((i = fread(s, 1, sizeof(s), f)) > 0) {
	    if ((i = fwrite(s, 1, i, stdout)) <= 0) {
		i = -1;
		break;
	    }
	}
	fclose(f);
    }
    fflush(stdout);
    remove(tmp);
    return i;
}

/*************************************************************************/

int main(int ac, char **av)
{
    static char str[1024], s[1024], t[1024];
    char *u, *v, c;
    char map[32];
    int scale, i, screen = 0, layer = 0, gif = 0;
    struct location *loc;

    if (!(u = getenv("QUERY_STRING")))
	doerror("Can't find request info");
    strncpy(s, u, 1023); s[1023] = 0;
    *str = 0;
    while (u = strchr(s, '%')) {
	*u++ = 0;
	strcat(str, s);
	*++u = dohexnum(u);
	strcpy(s, u);
    }
    strcat(str, s);
    sprintf(s, "&%s", str);

    while (u = strrchr(s, '&')) {
	*u++ = 0;
	strcpy(t, u);
	u = strchr(t, '=');
	if (!u) {
	    doerror("Bad argument line (missing '=')");
	}
	*u++ = 0;
	if (!strcmp(t, "loc"))
	    strscpy(map, u, sizeof(map));
	else if (!strcmp(t, "scale"))
	    scale = atoi(u);
	else if (!strcmp(t, "gif"))
	    gif = atoi(u);
	else {
	    sprintf(s, "Unknown parameter %s", t);
	    doerror(s);
	}
    }

    if (!*map || !scale) {
	doerror("Missing parameter(s)");
    }

    if (u = strchr(map, '/')) {
	if (v = strchr(u+1, '/')) {
	    *v++ = 0;
	    layer = atoi(v);
	}
	*u++ = 0;
	screen = atoi(u);
    }

    loc = find_location(map);
    if (!loc)
	doerror("Invalid location");
    else if (screen<0 || screen>=lenof(loc->floors) || !loc->floors[screen])
	doerror("Invalid floor");

    if (gif)
	i = send_gif(loc, screen, layer, scale);
    else
	i = send_html(loc, screen, layer, scale);

    exit(i<0 ? 20 : 0);
}
