/* Given one (possibly two) location(s), output an HTML document allowing
 * the user to pick the location, floor, and scale. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define strscpy(d,s,n) do { strncpy(d, s, n); d[n] = 0; } while (0)

/* Path to CGI programs */
#define CGI_PATH "/cgi-bin"


/* An error occurred - print an error message and quit */
void doerror(char *s)
{
    char *t;

    printf("Content-Type: text/html\n\n"
		"<html>\n<head>\n<title>Error</title>\n</head>\n"
		"<body>\n<h1>Error!</h1>\n"
		"Something has gone horribly wrong...<p>\n"
		"%s<p>\n</body>\n</html>\n", s);
    exit(20);
}

/* Translate a 2-digit hex number to a character */
char dohexnum(char *s)
{
    char t[3];
    strncpy(t, s, 2); t[2] = 0;

    if (!isxdigit(s[0]) || !isxdigit(s[1])) {
	doerror("Bad hex digit");
    }
    return strtol(t, NULL, 16);
}


main(int ac, char **av)
{
    static char s[1024], t[1024];
    char *u, *v, *w;
    char loc[32], loc2[32], name[32], name2[32], floors[512], floors2[512];
    int i;

    *loc = *loc2 = *floors = *floors2 = 0;

    if (!(u = getenv("QUERY_STRING")))
	doerror("Can't find request info");
    *(w=s) = '&'; *++w = 0;
    while (v = strchr(u, '%')) {
	strscpy(w, u, v-u);
	w += v-u;
	*w++ = dohexnum(v+1);
	u = v+3;
    }
    strcpy(w, u);

    while (v = strrchr(s, '&')) {
	*v++ = 0;
	strcpy(t, v);
	v = strchr(t, '=');
	if (!v) {
	    doerror("Bad argument line (missing '=')");
	}
	*v++ = 0;
	if (!strcmp(t, "loc"))
	    strscpy(loc, v, sizeof(loc)-1);
	else if (!strcmp(t, "loc2"))
	    strscpy(loc2, v, sizeof(loc2)-1);
	else if (!strcmp(t, "name"))
	    strscpy(name, v, sizeof(name)-1);
	else if (!strcmp(t, "name2"))
	    strscpy(name2, v, sizeof(name2)-1);
	else if (!strcmp(t, "floors"))
	    strscpy(floors, v, sizeof(floors)-1);
	else if (!strcmp(t, "floors2"))
	    strscpy(floors2, v, sizeof(floors2)-1);
	else {
	    sprintf(s, "Unknown parameter %s", t);
	    doerror(s);
	}
    }

    if (!*loc || !*name || (*loc2 && !*name2)) {
	doerror("Missing parameter(s)");
    }

    printf("\
Content-Type: text/html\n\
\n\
<html>\n\
<head>\n\
<title>%s%s%s</title>\n\
</head>\n\
\n\
<body>\n\
<form method=get action=\"%s/getmap.cgi\">\n",
	name, *loc2 ? " and " : "", *loc2 ? name2 : "", CGI_PATH);

    if (*loc2) {
	printf("Select a location:<p>\n<ul>\n");
	if (*floors) {
	    printf("<li>%s:\n<ul>\n", name);
	    i = 0; u = strtok(floors, "/");
	    do {
		printf("<li><input type=radio name=\"loc\" value=\"%s/%d\">"
			" %s</a>\n", loc, i, u);
		++i;
	    } while (u = strtok(NULL, "/"));
	    printf("</ul>\n");
	} else {
	    printf("<li><input type=radio name=\"loc\" value=\"%s\"> %s</a>\n",
								loc, name);
	}
	if (*floors2) {
	    printf("<li>%s:\n<ul>\n", name2);
	    i = 0; u = strtok(floors2, "/");
	    do {
		printf("<li><input type=radio name=\"loc\" value=\"%s/%d\">"
			" %s</a>\n", loc2, i, u);
		++i;
	    } while (u = strtok(NULL, "/"));
	    printf("</ul>\n");
	} else {
	    printf("<li><input type=radio name=\"loc\" value=\"%s\"> %s</a>\n",
								loc2, name2);
	}
	printf("</ul>\n");
    } else {
	if (*floors) {
	    printf("Selected location: %s<p>\n", name);
	    printf("Select a floor:<p>\n<ul>\n", name);
	    i = 0; u = strtok(floors, "/");
	    do {
		printf("<li><input type=radio name=\"loc\" value=\"%s/%d\">"
			" %s</a>\n", loc, i, u);
		++i;
	    } while (u = strtok(NULL, "/"));
	    printf("</ul>\n");
	} else {
	    printf("Selected location:<p>\n<ul>\n<li><input type=radio checked "
			"name=\"loc\" value=\"%s\"> %s\n</ul>\n", loc, name);
	}
    }

    printf("\
\n\
<hr width=50%>\n\
\n\
Select a scale factor:<p>\n\
<ul>\n\
<li><input type=radio name=\"scale\" value=1 checked> 1 : 1\n\
<li><input type=radio name=\"scale\" value=2> 1 : 2\n\
<li><input type=radio name=\"scale\" value=4> 1 : 4\n\
<li><input type=radio name=\"scale\" value=8> 1 : 8\n\
</ul>\n\
\n\
<input type=submit value=\"Get map\">\n\
\n\
</form>\n\
\n\
<hr>\n\
\n\
<center>Final Fantasy is copyright (c) Square Co., Ltd.</center>\n\
\n\
</body>\n\
</html>\n");

    exit(0);
}
